@@ -39,13 +39,17 @@ use IO::Handle;
 use lib '/usr/share/devscripts';
 use Devscripts::DB_File_Lock;
 use Fcntl qw(O_RDWR O_RDONLY O_CREAT F_SETFD);
+use Getopt::Long;
+
+# Funny UTF-8 warning messages from HTML::Parse should be ignorable (#292671)
+$SIG{'__WARN__'} = sub { warn $_[0] unless $_[0] =~ /^Parsing of undecoded UTF-8 will give garbage when decoding entities/; };
 
 my $it = undef;
 my $lwp_broken = undef;
 my $ua;
 
 sub have_lwp() {
-    return $lwp_broken if defined $lwp_broken;
+    return ($lwp_broken ? 0 : 1) if defined $lwp_broken;
     eval {
 	require LWP;
 	require LWP::UserAgent;
@@ -178,20 +182,35 @@ information on setting up a cache.
 Opposite of --online; overrides any configuration file directive to work
 offline.
 
-=item -m, --full-mirror
+=item --cache, --no-cache
 
-When running a B<bts cache> command, also mirror the boring
-attachments to the BTS bug pages, including the mbox version of the
-bug report and the acknowledgement emails.
+Should we attempt to to cache new versions of BTS pages when
+performing show/bugs commands?  Default is to cache.
 
-=item --no-full-mirror
+=item --cache-mode={min|mbox|full}
 
-Override --full-mirror.
+When running a B<bts cache> command, should we only mirror the basic
+bug (min), or should we also mirror the mbox version (mbox), or should
+we mirror the whole thing, including the mbox and the boring
+attachments to the BTS bug pages and the acknowledgement emails (full)?
+Default is min.
+
+=item --mbox
+
+Open a mail reader to read the mbox corresponding to a given bug number
+for show and bugs commands.
+
+=item --mailreader=READER
+
+Specify the command to read the mbox.  Must contain a "%s" string, which
+will be replaced by the name of the mbox file.  The command will be split
+on white space and will not be passed to a shell.  Default is 'mutt -f %s'.
+(Also, %% will be substituted by a single % if this is needed.)
 
 =item -f, --force-refresh
 
 Download a bug report again, even if it does not appear to have
-changed since the last cache command.  Useful if a --full-mirror is
+changed since the last cache command.  Useful if a --cache-mode=full is
 requested for the first time (otherwise unchanged bug reports will not
 be downloaded again, even if the boring bits have not been
 downloaded).
@@ -200,6 +219,12 @@ downloaded).
 
 Suppress any configuration file --force-refresh option.
 
+=item -q, --quiet
+
+When running bts cache, only display information about newly cached
+pages, not messages saying already cached.  If this option is
+specified twice, only output error messages (to stderr).
+
 =item --no-conf, --noconf
 
 Do not read any configuration files.  This can only be used as the
@@ -212,9 +237,10 @@ first option given on the command-line.
 # Start by setting default values
 
 my $offlinemode=0;
-my $cachemode=1;
-my $fullmirrormode=0;
+my $caching=1;
+my $cachemode='min';
 my $refreshmode=0;
+my $mailreader='mutt -f %s';
 
 # Next, read read configuration files and then command line
 # The next stuff is boilerplate
@@ -227,15 +253,16 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     my %config_vars = (
 		       'BTS_OFFLINE' => 'no',
 		       'BTS_CACHE' => 'yes',
-		       'BTS_FULL_MIRROR' => 'no',
+		       'BTS_CACHE_MODE' => 'min',
 		       'BTS_FORCE_REFRESH' => 'no',
+		       'BTS_MAIL_READER' => 'mutt -f %s',
 		       );
     my %config_default = %config_vars;
     
     my $shell_cmd;
     # Set defaults
     foreach my $var (keys %config_vars) {
-	$shell_cmd .= "$var='$config_vars{$var}';\n";
+	$shell_cmd .= qq[$var="$config_vars{$var}";\n];
     }
     $shell_cmd .= 'for file in ' . join(" ",@config_files) . "; do\n";
     $shell_cmd .= '[ -f $file ] && . $file; done;' . "\n";
@@ -249,10 +276,12 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'BTS_OFFLINE'}='no';
     $config_vars{'BTS_CACHE'} =~ /^(yes|no)$/
 	or $config_vars{'BTS_CACHE'}='yes';
-    $config_vars{'BTS_FULL_MIRROR'} =~ /^(yes|no)$/
-	or $config_vars{'BTS_FULL_MIRROR'}='no';
+    $config_vars{'BTS_CACHE_MODE'} =~ /^(min|mbox|full)$/
+	or $config_vars{'BTS_CACHE_MODE'}='min';
     $config_vars{'BTS_FORCE_REFRESH'} =~ /^(yes|no)$/
 	or $config_vars{'BTS_FORCE_REFRESH'}='no';
+    $config_vars{'BTS_MAIL_READER'} =~ /\%s/
+	or $config_vars{'BTS_MAIL_READER'}='mutt -f %s';
 
     foreach my $var (sort keys %config_vars) {
 	if ($config_vars{$var} ne $config_default{$var}) {
@@ -263,74 +292,77 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     chomp $modified_conf_msg;
 
     $offlinemode = $config_vars{'BTS_OFFLINE'} eq 'yes' ? 1 : 0;
-    $cachemode = $config_vars{'BTS_CACHE'} eq 'no' ? 0 : 1;
-    $fullmirrormode = $config_vars{'BTS_FULL_MIRROR'} eq 'yes' ? 1 : 0;
+    $caching = $config_vars{'BTS_CACHE'} eq 'no' ? 0 : 1;
+    $cachemode = $config_vars{'BTS_CACHE_MODE'};
     $refreshmode = $config_vars{'BTS_FORCE_REFRESH'} eq 'yes' ? 1 : 0;
+    $mailreader = $config_vars{'BTS_MAIL_READER'};
 }
 
 if (exists $ENV{'BUGSOFFLINE'}) {
     warn "BUGSOFFLINE environment variable deprecated: please use ~/.devscripts\nor --offline/-o option instead!  (See bts(1) for details.)\n";
 }
 
-# For now, a very simple parser, instead of Getopt::Long since there are
-# so few options.
-while (@ARGV) {
-    my $arg=$ARGV[0];
-    if ($arg =~ /^--?(.*)/) {
-	my $option=$1;
-	shift @ARGV;
-	if ($option eq 'offline' || $option eq 'o') {
-	    $offlinemode=1;
-	}
-	elsif ($option eq 'online' || $option =~ /^no-?offline$/) {
-	    $offlinemode=0;
-	}
-	if ($option eq 'cache') {
-	    $cachemode=1;
-	}
-	elsif ($option =~ /^no-?cache$/) {
-	    $cachemode=0;
-	}
-	elsif ($option eq 'full-mirror' || $option eq 'm') {
-	    $fullmirrormode=1;
-	}
-	elsif ($option =~ /^no-?full-mirror$/) {
-	    $fullmirrormode=1;
-	}
-	elsif ($option eq 'force-refresh' || $option eq 'f') {
-	    $refreshmode=1;
-	}
-	elsif ($option =~ /^no-?force-refresh$/) {
-	    $refreshmode=1;
-	}
-	elsif ($option eq 'help' || $option eq 'h') {
-	    bts_help(); exit 0;
-	}
-	elsif ($option eq 'version' || $option eq 'v') {
-	    bts_version(); exit 0;
-	}
-	else {
-	    die "$progname: Unknown option, \"$arg\"\nRun $progname --help for more information\n";
-	}
-    }
-    else {
-	last; # end of options
+my ($opt_help, $opt_version, $opt_noconf);
+my ($opt_cachemode, $opt_mailreader);
+my $mboxmode = 0;
+my $quiet=0;
+
+Getopt::Long::Configure('require_order');
+GetOptions("help|h" => \$opt_help,
+	   "version" => \$opt_version,
+	   "o" => \$offlinemode,
+	   "offline!" => \$offlinemode,
+	   "online" => sub { $offlinemode = 0; },
+	   "cache!" => \$caching,
+	   "cache-mode|cachemode=s" => \$opt_cachemode,
+	   "m|mbox" => \$mboxmode,
+	   "mailreader|mail-reader=s" => \$opt_mailreader,
+	   "f" => \$refreshmode,
+	   "force-refresh!" => \$refreshmode,
+	   "q|quiet+" => \$quiet,
+	   "noconf|no-conf" => \$opt_noconf,
+	   )
+    or die "Usage: bts [options]\nRun $progname --help for more details\n";
+
+if ($opt_noconf) {
+    die "$progname: --no-conf is only acceptable as the first command-line option!\n";
+}
+if ($opt_help) { bts_help(); exit 0; }
+if ($opt_version) { bts_version(); exit 0; }
+
+if ($opt_mailreader) {
+    if ($opt_mailreader =~ /\%s/) {
+	$mailreader=$opt_mailreader;
+    } else {
+	warn "bts: ignoring invalid --mailreader option: invalid mail command following it.\n";
     }
 }
 
+if ($opt_cachemode) {
+    if ($opt_cachemode =~ /^(min|mbox|full)$/) {
+	$cachemode=$opt_cachemode;
+    } else {
+	warn "bts: ignoring invalid --cache-mode; must be one of min, mbox, full.\n";
+    }
+}
+
+
 if (@ARGV == 0) {
     bts_help();
     exit 0;
 }
 
+
 # Otherwise, parse the arguments
 my @command;
 my @args;
 our @comment=('');
 my $ncommand = 0;
 my $iscommand = 1;
-foreach (@ARGV) {
+while (@ARGV) {
+    $_ = shift @ARGV;
     if ($_ =~ /^[\.,]$/) {
+	next if $iscommand;  # ". ." in command line - oops!
 	$ncommand++;
 	$iscommand = 1;
 	$comment[$ncommand] = '';
@@ -346,7 +378,7 @@ foreach (@ARGV) {
 	push @{$args[$ncommand]}, $_;
     }
 }
-push @command, '' if $iscommand;
+$ncommand-- if $iscommand;
 
 # Grub through the symbol table to find matching commands.
 my $subject = '';
@@ -369,32 +401,93 @@ exit 0;
 
 =head1 COMMANDS
 
-For full details about the commands, see the BTS documentation.
+For full details about the commands, see the L<bts(1)> manpage.
 
 =over 4
 
-=item show <bug>
+=item show [options] [<bug number> | <package> | <maintainer> | : ]
 
-Display a particular bug in a web browser using
-L<sensible-browser(1)>.  If either the -o or --offline option is used,
-or there is already an up-to-date copy in the local cache, the cached
-version will be used.
+=item show [options] [src:<package> | from:<submitter> | tag:<tag> ]
 
-Also, if caching has been enabled (that is, there exists a cache
-directory ~/.devscripts_cache/bts/), then any page requested by "bts
-show" will automatically be cached, and therefore available offline
-thereafter.  Pages which are automatically cached in this way will be
-deleted on subsequent "bts show|bugs|cache" invocations if they have
-not been accessed in 30 days.  This automatic caching can be disabled
-using the --no-cache option or explicitly enabled using the --cache
-option.
+This is a synonym for bts bugs.
+
+=cut
+
+sub bts_show {
+    goto &bts_bugs;
+}
+
+=item bugs [options] [<bug number> | <package> | <maintainer> | : ]
+
+=item bugs [options] [src:<package> | from:<submitter> | tag:<tag> ]
+
+Display the page listing the requested bugs in a web browser using
+L<sensible-browser(1)>.
+
+Options may be specified after the "bugs" command in addition to or
+instead of options at the start of the command line: recognised
+options at his point are: -o/--offline/--online, --mbox, --mailreader
+and --[no-]cache.  These are described earlier in this manpage.  If
+either the -o or --offline option is used, or there is already an
+up-to-date copy in the local cache, the cached version will be used.
+
+The meanings of the possible arguments are as follows:
+
+=over 8
+
+=item (none)
+
+If nothing is specified, bts bugs will display your bugs, assuming
+that either DEBEMAIL or EMAIL (examined in that order) is set to the
+appropriate email address.
+
+=item <bug number>
+
+Display bug number <bug number>.
+
+=item <package>
+
+Display the bugs for the package <package>.
+
+=item src:<package>
+
+Display the bugs for the source package <package>.
+
+=item <maintainer>
+
+Display the bugs for the maintainer email address <maintainer>.
+
+=item from:<submitter>
+
+Display the bugs for the submitter <submitter>.
+
+=item tag:<tag>
+
+Display the bugs which are tagged with <tag>.
+
+=item :
+
+Details of the bug tracking system itself, along with a bug-request
+page with more options than this script, can be found on
+http://bugs.debian.org/.  This page itself will be opened if the
+command 'bts bugs :' is used.
+
+=back
+
+If caching has been enabled (that is, there exists a cache directory
+~/.devscripts_cache/bts/), then any page requested by "bts show" will
+automatically be cached, and therefore available offline thereafter.
+Pages which are automatically cached in this way will be deleted on
+subsequent "bts show|bugs|cache" invocations if they have not been
+accessed in 30 days.  This automatic caching can be disabled using the
+--no-cache option or explicitly enabled using the --cache option.
 
 Any other B<bts> commands following this on the command line will be
 executed after the browser has been exited.
 
 The desired browser can be specified and configured by setting the
 BROWSER environment variable.  The conventions follow those defined by
-Eric Raymond at http://catb.org/~esr/BROWSER/; we here reproduce the
+Eric Raymond at http://www.catb.org/~esr/BROWSER/; we here reproduce the
 relevant part.
 
 The value of BROWSER may consist of a colon-separated series of
@@ -419,65 +512,37 @@ BROWSER='mozilla -raise -remote "openURL(%s,new-window)":links'
 
 =cut
 
-sub bts_show {
-    my $thing=shift or die "bts show: display what bug?\n";
-    browse($thing);
-}
-
-=item bugs [<package> | src:<package> | <maintainer> | from:<submitter>]
-
-=item bugs [tag:<tag> | <number> | : ]
-
-Display the page listing the requested bugs in a web browser.  The
-meanings of the possible arguments are as follows:
-
-=over 8
-
-=item (none)
-
-If nothing is specified, bts bugs will display your bugs, assuming
-that either DEBEMAIL or EMAIL (examined in that order) is set to the
-appropriate email address.
-
-=item <package>
-
-Display the bugs for the package <package>.
-
-=item src:<package>
-
-Display the bugs for the source package <package>.
-
-=item <maintainer>
-
-Display the bugs for the maintainer email address <maintainer>.
-
-=item from:<submitter>
-
-Display the bugs for the submitter <submitter>.
-
-=item tag:<tag>
-
-Display the bugs which are tagged with <tag>.
-
-=item <number>
-
-Display bug number <number>.
-
-=item :
-
-Details of the bug tracking system itself, along with a bug-request
-page with more options than this script, can be found on
-http://bugs.debian.org/.  This page itself will be opened if the
-command 'bts bugs :' is used.
-
-=back
-
-All of the other comments above about "bts show" apply equally to "bts
-bugs".
-
-=cut
-
 sub bts_bugs {
+    @ARGV = @_;
+    my ($sub_offlinemode, $sub_caching, $sub_mboxmode, $sub_mailreader);
+    GetOptions("o" => \$sub_offlinemode,
+	       "offline!" => \$sub_offlinemode,
+	       "online" => sub { $sub_offlinemode = 0; },
+	       "cache!" => \$sub_caching,
+	       "m|mbox" => \$sub_mboxmode,
+	       "mailreader|mail-reader=s" => \$sub_mailreader,
+	       )
+    or die "bts: unknown options for bugs command\n";
+    @_ = @ARGV; # whatever's left
+
+    if (defined $sub_offlinemode) {
+	($offlinemode, $sub_offlinemode) = ($sub_offlinemode, $offlinemode);
+    }
+    if (defined $sub_caching) {
+	($caching, $sub_caching) = ($sub_caching, $caching);
+    }
+    if (defined $sub_mboxmode) {
+	($mboxmode, $sub_mboxmode) = ($sub_mboxmode, $mboxmode);
+    }
+    if (defined $sub_mailreader) {
+	if ($sub_mailreader =~ /\%s/) {
+	    ($mailreader, $sub_mailreader) = ($sub_mailreader, $mailreader);
+	} else {
+	    warn "bts: ignoring invalid --mailreader $sub_mailreader option:\ninvalid mail command following it.\n";
+	    $sub_mailreader = undef;
+	}
+    }
+
     my $url = shift;
     if (! $url) {
 	if (defined $ENV{'DEBEMAIL'}) {
@@ -493,6 +558,20 @@ sub bts_bugs {
     if ($url =~ /^.*\s+<(.*)>$/) { $url = $1; }
     $url =~ s/^:$//;
     browse($url);
+
+    # revert options
+    if (defined $sub_offlinemode) {
+	$offlinemode = $sub_offlinemode;
+    }
+    if (defined $sub_caching) {
+	$caching = $sub_caching;
+    }
+    if (defined $sub_mboxmode) {
+	$mboxmode = $sub_mboxmode;
+    }
+    if (defined $sub_mailreader) {
+	$mailreader = $sub_mailreader;
+    }
 }
 
 =item clone <bug> [new IDs]
@@ -754,7 +833,7 @@ sub bts_noowner {
     mailbts("bug $bug has no owner", "noowner $bug");
 }
 
-=item cache [<maint email> | <pkg> | src:<pkg> | from:<submitter>]
+=item cache [options] [<maint email> | <pkg> | src:<pkg> | from:<submitter>]
 
 Generate or update a cache of bug reports for the given email address
 or package. By default it downloads all bugs belonging to the email
@@ -780,18 +859,49 @@ for bugs in the cache even when you're online, as it can just compare the
 item in the cache with what's on the server, and not re-download it
 every time.
 
-Two options affect the behaviour of the cache command.  The first is
---full-mirror, which forces B<bts> to download all of the referenced
-links from the bug page, including boring bits such as the
-acknowledgement emails, emails to the control bot, and the mbox
-version of the bug report.  This can be switched off using
---no-full-mirror.  The second is --force-refresh, which forces the
-download, even if the cached bug report is up-to-date.  Both of these
-are configurable from the configuration file, as described below.
+Some options affect the behaviour of the cache command.  The first is
+the setting of --cache-mode, which controls how much B<bts> downloads
+of the referenced links from the bug page, including boring bits such
+as the acknowledgement emails, emails to the control bot, and the mbox
+version of the bug report.  It can take three values: min (the
+minimum), mbox (download the minimum plus the mbox version of the bug
+report) or full (the whole works).  The second is --force-refresh or
+-f, which forces the download, even if the cached bug report is
+up-to-date.  Both of these are configurable from the configuration
+file, as described below.  They may also be specified after the
+"cache" command as well as at the start of the command line.
+
+Finally, -q or --quiet will suppress messages about caches being
+up-to-date, and giving the option twice will suppress all cache
+messages (except for error messages).
 
 =cut
 
 sub bts_cache {
+    @ARGV = @_;
+    my ($sub_cachemode, $sub_refreshmode);
+    my $sub_quiet = $quiet;
+    GetOptions("cache-mode|cachemode=s" => \$sub_cachemode,
+	       "f" => \$sub_refreshmode,
+	       "force-refresh!" => \$sub_refreshmode,
+	       "q|quiet+" => \$sub_quiet,
+	       )
+    or die "bts: unknown options for bugs command\n";
+    @_ = @ARGV; # whatever's left
+
+    if (defined $sub_refreshmode) {
+	($refreshmode, $sub_refreshmode) = ($sub_refreshmode, $refreshmode);
+    }
+    if (defined $sub_cachemode) {
+	if ($sub_cachemode =~ /^(min|mbox|full)$/) {
+	    ($cachemode, $sub_cachemode) = ($sub_cachemode, $cachemode);
+	} else {
+	    warn "bts: ignoring invalid --cache-mode $sub_cachemode;\nmust be one of min, mbox, full.\n";
+	}
+    }
+    # This may be a no-op, we don't mind
+    ($quiet, $sub_quiet) = ($sub_quiet, $quiet);
+
     prunecache();
     if (! have_lwp()) {
 	die "Couldn't run bts cache: $lwp_broken\n";
@@ -800,14 +910,14 @@ sub bts_cache {
     if (! -d $cachedir) {
 	if (! -d dirname($cachedir)) {
 	    mkdir(dirname($cachedir))
-		or die "bts: couldn't mkdir ".dirname($cachedir).": $!";
+		or die "bts: couldn't mkdir ".dirname($cachedir).": $!\n";
 	}
 	mkdir($cachedir)
-	    or die "bts: couldn't mkdir $cachedir: $!";
+	    or die "bts: couldn't mkdir $cachedir: $!\n";
     }
     
     my $tocache;
-    if ($#_ != -1) { $tocache=shift; }
+    if (@_ > 0) { $tocache=shift; }
     else { $tocache=''; }
     
     if (! length $tocache) {
@@ -829,7 +939,7 @@ sub bts_cache {
     if (@oldbugs) {
 	tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	     O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
-	    or die "bts: couldn't open DB file $timestampdb for writing: $!"
+	    or die "bts: couldn't open DB file $timestampdb for writing: $!\n"
 	    if ! tied %timestamp;
     }
 
@@ -845,6 +955,15 @@ sub bts_cache {
     foreach my $bug (keys %bugs) {
 	download($bug, 1);
     }
+
+    # revert options    
+    if (defined $sub_refreshmode) {
+	$refreshmode = $sub_refreshmode;
+    }
+    if (defined $sub_cachemode) {
+	$cachemode = $sub_cachemode;
+    }
+    $quiet = $sub_quiet;
 }
 
 =item cleancache <package> | src:<package> | <maintainer>
@@ -879,7 +998,7 @@ sub bts_cleancache {
     # clean index
     tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	 O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
-	or die "bts: couldn't open DB file $timestampdb for writing: $!"
+	or die "bts: couldn't open DB file $timestampdb for writing: $!\n"
 	if ! tied %timestamp;
 
     if ($toclean =~ /^\d+$/) {
@@ -941,8 +1060,14 @@ Valid options are:
                           pages when performing show/bug commands
    --cache                Do attempt to cache new versions of BTS
                           pages when performing show/bug commands (default)
-   -m, --full-mirror      Download boring bits when running cache command
-   --no-full-mirror       Don\'t download the boring bits (default)
+   --cache-mode={min|mbox|full}
+                          How much to cache when we\'re caching: the sensible
+                          bare minimum (default), the mbox as well, or
+                          everything?
+   -m, --mbox             With show or bugs, open a mailreader to read the mbox
+                          version instead
+   --mailreader=CMD       Run CMD to read an mbox; default is 'mutt -f %s'
+                          (must contain %s, which is replaced by mbox name)
    -f, --force-refresh    Reload all bug reports being cached, even unchanged
                           ones
    --no-force-refresh     Don\'t do so (default)
@@ -983,6 +1108,7 @@ sub checkbug {
     }
     
     $bug=~s/^[^-0-9]*//;
+    $bug=~s/:$//;
     if (! exists $clonedbugs{$bug} &&
 	(! length $bug || $bug !~ /^[0-9]+$/)) {
 	warn "\"$_[0]\" does not look like a bug number\n";
@@ -1039,7 +1165,7 @@ sub mailbtsall {
 
 	my $pid = open(MAIL, "|-");
 	if (! defined $pid) {
-	    die "bts: Couldn't fork: $!";
+	    die "bts: Couldn't fork: $!\n";
 	}
 	if ($pid) {
 	    # parent
@@ -1053,16 +1179,16 @@ X-BTS-Version: $version
 # Automatically generated email from bts, devscripts version $version
 $body
 EOM
-	    close MAIL or die "bts: sendmail error: $!";
+	    close MAIL or die "bts: sendmail error: $!\n";
 	}
 	else {
 	    # child
 	    if ($debug) {
 		exec("/bin/cat")
-		    or die "bts: error running cat: $!";
+		    or die "bts: error running cat: $!\n";
 	    } else {
 		exec("/usr/sbin/sendmail", "-t")
-		    or die "bts: error running sendmail: $!";
+		    or die "bts: error running sendmail: $!\n";
 	    }
 	}
     }
@@ -1074,12 +1200,12 @@ EOM
 	if ($pid) {
 	    # parent
 	    print MAIL $body;
-	    close MAIL or die "bts: mail: $!";
+	    close MAIL or die "bts: mail: $!\n";
 	}
 	else {
 	    # child
 	    exec("mail", "-s$subject", $btsemail)
-		or die "bts: error running mail: $!";
+		or die "bts: error running mail: $!\n";
 	}
     }
 }
@@ -1092,6 +1218,7 @@ EOM
 sub download {
     my $thing=shift;
     my $manual=shift;  # true="bts cache", false="bts show/bug"
+    my $mboxing=shift;  # true="bts --mbox show/bugs", and only if $manual=0
     my $timestamp = 0;
     my $url;
 
@@ -1099,10 +1226,10 @@ sub download {
     $url = "$btsurl$thing";
 
     if (! -d $cachedir) {
-	die "bts: download() called but no cachedir!";
+	die "bts: download() called but no cachedir!\n";
     }
 
-    chdir($cachedir) || die "bts: chdir $cachedir: $!";
+    chdir($cachedir) || die "bts: chdir $cachedir: $!\n";
 
     if (-f cachefile($thing)) {
 	$timestamp = get_timestamp($thing) || 0;
@@ -1110,9 +1237,24 @@ sub download {
 	if (is_manual($timestamp)) { $manual = 1; }
     }
 
-    print "Downloading $url ... ";
+    # do we actually have to do more than we might have thought?
+    # yes, if we've caching with --cache-mode=mbox or full and the bug had
+    # previously been cached in a less thorough format
+    my $forcedownload = 0;
+    if ($thing =~ /^\d+$/ and ! $refreshmode and $cachemode ne 'min') {
+	if (! -r mboxfile($thing)) {
+	    $forcedownload = 1;
+	} elsif ($cachemode eq 'full' and -d $thing) {
+	    opendir DIR, $thing or die "bts: opendir $cachedir/$thing: $!\n";
+	    my @htmlfiles = grep { /^\d+\.html$/ } readdir(DIR);
+	    closedir DIR;
+	    $forcedownload = 1 unless @htmlfiles;
+	}
+    }
+
+    print "Downloading $url ... " if ! $quiet;
     IO::Handle::flush(\*STDOUT);
-    my ($ret, $msg, $livepage) = bts_mirror($url, $timestamp);
+    my ($ret, $msg, $livepage) = bts_mirror($url, $timestamp, $forcedownload);
     if ($ret == MIRROR_UP_TO_DATE) {
 	# we have an up-to-date version already, nothing to do
 	# and $timestamp is guaranteed to be well-defined
@@ -1120,7 +1262,7 @@ sub download {
 	    set_timestamp($thing, make_manual($timestamp));
 	}
 
-	print "(cache already up-to-date)\n";
+	print "(cache already up-to-date)\n" if ! $quiet;
 	return "";
     }
     elsif ($ret == MIRROR_DOWNLOADED) {
@@ -1128,7 +1270,7 @@ sub download {
 	# we've successfully stashed the data away
 	$timestamp = time;
 
-	die "bts: empty page downloaded" unless length $livepage;
+	die "bts: empty page downloaded\n" unless length $livepage;
 
 	my $bug2filename = { };
 
@@ -1141,16 +1283,22 @@ sub download {
 
 	my $data = $livepage;  # work on a copy, not the original
 	my $cachefile=cachefile($thing);
-	open (OUT_CACHE, ">$cachefile") or die "bts: open $cachefile: $!";
+	open (OUT_CACHE, ">$cachefile") or die "bts: open $cachefile: $!\n";
 
 	$data = mangle_cache_file($data, $thing, $bug2filename, $timestamp);
 	print OUT_CACHE $data;
-	close OUT_CACHE or die "bts: problems writing to $cachefile: $!";
+	close OUT_CACHE or die "bts: problems writing to $cachefile: $!\n";
 
 	set_timestamp($thing,
 	    $manual ? make_manual($timestamp) : make_automatic($timestamp));
 
-	print "(cached new version)\n";
+	if ($quiet == 0) {
+	    print "(cached new version)\n";
+	} elsif ($quiet == 1) {
+	    print "Downloading $url ... (cached new version)\n";
+	} elsif ($quiet > 1) {
+	    # do nothing
+	}
 	return $livepage;
     } else {
 	die "bts: couldn't download $url:\n$msg\n";
@@ -1194,7 +1342,7 @@ sub download_attachments {
 	    # already downloaded?
 	    next if -f $bug2filename{$msg} and not $refreshmode;
 	}
-	elsif ($fullmirrormode and
+	elsif ($cachemode eq 'full' and
 	       m%<a href="(bugreport\.cgi\?bug=\d+&amp;msg=(\d+))">%) {
 	    $ref = $1;
 	    $msg = $2;
@@ -1202,11 +1350,11 @@ sub download_attachments {
             # already downloaded?
 	    next if -f $bug2filename{$msg} and not $refreshmode;
 	}
-	elsif ($fullmirrormode and
+	elsif (($cachemode eq 'full' or $cachemode eq 'mbox' or $mboxmode) and
 	       m%<a href="(bugreport\.cgi\?bug=\d+&amp;mbox=yes)">%) {
 	    $ref = $1;
 	    $msg = 'mbox';
-	    $bug2filename{$msg} = "$thing/$msg";
+	    $bug2filename{$msg} = "$thing.mbox";
 	    # This always needs refreshing, as it does change as the bug
 	    # changes
 	}
@@ -1231,12 +1379,12 @@ sub download_attachments {
 		$data =~ s%<HEAD>%<HEAD><BASE href="../">%;
 		$data = mangle_cache_file($data, $thing,
 				          { $msg => "$thing/$msg.html",
-					    'mbox' => "$thing/mbox", },
+					    'mbox' => "$thing.mbox", },
 					  $timestamp);
 	    }
 	    mkpath(dirname $bug2filename{$msg});
 	    open OUT_CACHE, ">$bug2filename{$msg}"
-	        or die "bts: open cache $bug2filename{$msg}";
+	        or die "bts: open cache $bug2filename{$msg}\n";
 	    print OUT_CACHE $data;
 	    close OUT_CACHE;
 	} else {
@@ -1249,6 +1397,54 @@ sub download_attachments {
 }
 
 
+# Download the mailbox for a given bug, return mbox ($fh, filename) on success,
+# die on failure
+sub download_mbox {
+    my $thing = shift;
+    my $temp = shift;  # do we wish to store it in cache or in a temp file?
+    my $mboxfile = mboxfile($thing);
+
+    die "bts: trying to download mbox for illegal bug number $thing.\n"
+	unless $mboxfile;
+
+    if (! have_lwp()) {
+	die "bts: couldn't run bts --mbox: $lwp_broken\n";
+    }
+    init_agent() unless $ua;
+
+    my $request = HTTP::Request->new('GET', $btscgiurl . "bugreport.cgi?bug=$thing&mbox=yes");
+    my $response = $ua->request($request);
+    if ($response->is_success) {
+	my $content_length = defined $response->content ?
+	    length($response->content) : 0;
+	if ($content_length == 0) {
+	    die "bts: failed to download mbox.\n";
+	}
+
+	my ($fh, $filename);
+	if ($temp) {
+	    ($fh,$filename) = tempfile("btsXXXXXX",
+				       SUFFIX => ".mbox",
+				       DIR => File::Spec->tmpdir,
+				       UNLINK => 1);
+	    # Use filehandle for security
+	    open (OUT_MBOX, ">/dev/fd/" . fileno($fh))
+		or die "bts: writing to temporary file: $!\n";
+	} else {
+	    $filename = $mboxfile;
+	    open (OUT_MBOX, ">$mboxfile")
+		or die "bts: writing to mbox file $mboxfile: $!\n";
+	}
+	print OUT_MBOX $response->content;
+	close OUT_MBOX;
+	    
+	return ($fh, $filename);
+    } else {
+	die "bts: failed to download mbox.\n";
+    }
+}
+
+
 # Mangle downloaded file to work in the local cache, so
 # selectively modify the links
 sub mangle_cache_file {
@@ -1297,30 +1493,37 @@ sub deletecache {
     my $thing=shift;
 
     if (! -d $cachedir) {
-	die "bts: deletecache() called but no cachedir!";
+	die "bts: deletecache() called but no cachedir!\n";
     }
 
     delete_timestamp($thing);
     unlink cachefile($thing);
     if ($thing =~ /^\d+$/) {
 	rmtree("$cachedir/$thing", 0, 1) if -d "$cachedir/$thing";
+	unlink("$cachedir/$thing.mbox") if -f "$cachedir/$thing.mbox";
     }
 }
 
 # Given a thing, returns the filename for it in the cache.
 sub cachefile {
     my $thing=shift;
-    if ($thing eq '') { die "bts: cachefile given empty argument"; }
+    if ($thing eq '') { die "bts: cachefile given empty argument\n"; }
     $thing =~ s/^src:/src_/;
     $thing =~ s/^from:/from_/;
     $thing =~ s/^tag:/tag_/;
     return $cachedir.$thing.".html";
 }
 
+# Given a thing, returns the filename for its mbox in the cache.
+sub mboxfile {
+    my $thing=shift;
+    return $thing =~ /^\d+$/ ? $cachedir.$thing.".mbox" : undef;
+}
+
 # Given a bug number, returns the dirname for it in the cache.
 sub cachebugdir {
     my $thing=shift;
-    if ($thing !~ /^\d+$/) { die "bts: cachebugdir given faulty argument: $thing"; }
+    if ($thing !~ /^\d+$/) { die "bts: cachebugdir given faulty argument: $thing\n"; }
     return $cachedir.$thing;
 }
 
@@ -1342,7 +1545,7 @@ sub bugs_from_thing {
 
     if (-f $cachefile) {
 	local $/;
-	open (IN, $cachefile) || die "bts: open $cachefile: $!";
+	open (IN, $cachefile) || die "bts: open $cachefile: $!\n";
 	my $data=<IN>;
 	close IN;
 
@@ -1362,6 +1565,10 @@ sub browse {
 
     my $hascache=-d $cachedir;
     my $cachefile=cachefile($thing);
+    my $mboxfile=mboxfile($thing);
+    if ($mboxmode and ! $mboxfile) {
+	die "bts: you can only request a mailbox for a single bug report.\n";
+    }
 
     # Check that if we're requesting a tag, that it's a valid tag
     if ($thing =~ /^tag:(.*)$/) {
@@ -1374,35 +1581,53 @@ sub browse {
 	if (! $hascache) {
 	    die "bts: Sorry, you are in offline mode and have no cache. Run \"bts cache\" to create one.\n";
 	}
-	elsif (! -e $cachefile) {
-	    die "bts: Sorry, you are in offline mode and that is not cached. Use \"bts cache\" to update the cache.\n";
+	elsif ((! $mboxmode and ! -r $cachefile) or
+	       ($mboxmode and ! -r $mboxfile)) {
+	    die "bts: Sorry, you are in offline mode and that is not cached.\nUse \"bts [--cache-mode=...] cache\" to update the cache.\n";
 	}
-	runbrowser($cachefile);
-    }
-    # else we're in online mode
-    elsif ($hascache && $cachemode && have_lwp() && $thing ne '') {
-	my $live=download($thing);
-	
-	if (length($live)) {
-	    my ($fh,$livefile) = tempfile("btsXXXXXX",
-					  SUFFIX => ".html",
-					  DIR => File::Spec->tmpdir,
-					  UNLINK => 1);
-
-	    # Use filehandle for security
-	    open (OUT_LIVE, ">/dev/fd/" . fileno($fh))
-		or die "bts: writing to temporary file: $!";
-	    # Correct relative urls to point to the bts.
-	    $live =~ s/(?!\/)(\w+\.cgi)/$btscgiurl$1/g;
-	    print OUT_LIVE $live;
-	    # Some browsers don't like unseekable filehandles, so use filename
-	    runbrowser($livefile);
+	if ($mboxmode) {
+	    runmailreader($mboxfile);
 	} else {
 	    runbrowser($cachefile);
 	}
     }
+    # else we're in online mode
+    elsif ($hascache && $caching && have_lwp() && $thing ne '') {
+	my $live=download($thing, 0, $mboxmode);
+	
+	if ($mboxmode) {
+	    runmailreader($mboxfile);
+	} else {
+	    if (length($live)) {
+		my ($fh,$livefile) = tempfile("btsXXXXXX",
+					      SUFFIX => ".html",
+					      DIR => File::Spec->tmpdir,
+					      UNLINK => 1);
+
+		# Use filehandle for security
+		open (OUT_LIVE, ">/dev/fd/" . fileno($fh))
+		    or die "bts: writing to temporary file: $!\n";
+		# Correct relative urls to point to the bts.
+		$live =~ s/(?!\/)(\w+\.cgi)/$btscgiurl$1/g;
+		print OUT_LIVE $live;
+		# Some browsers don't like unseekable filehandles,
+		# so use filename
+		runbrowser($livefile);
+	    } else {
+		runbrowser($cachefile);
+	    }
+	}
+    }
     else {
-	runbrowser($btsurl.$thing);
+	if ($mboxmode) {
+	    # we appear not to be caching; OK, we'll download to a
+	    # temporary file
+	    warn "bts debug: downloading ${btscgiurl}bugreport.cgi?bug=$thing&mbox=yes\n" if $debug;
+	    my ($fh, $fn) = download_mbox($thing, 1);
+	    runmailreader($fn);
+	} else {
+	    runbrowser($btsurl.$thing);
+	}
     }
 }
 
@@ -1415,12 +1640,12 @@ sub prunecache {
     return unless -d $cachedir;
     return if -f $prunestamp and -M _ < 1;
 
-    chdir($cachedir) || die "bts: chdir $cachedir: $!";
+    chdir($cachedir) || die "bts: chdir $cachedir: $!\n";
 
     # remove the now-defunct live-download file
     unlink "live_download.html";
 
-    opendir DIR, '.' or die "bts: opendir $cachedir: $!";
+    opendir DIR, '.' or die "bts: opendir $cachedir: $!\n";
     my @cachefiles = grep { ! /^\.\.?$/ } readdir(DIR);
     closedir DIR;
 
@@ -1436,6 +1661,8 @@ sub prunecache {
     foreach (@cachefiles) {
 	if (/^(\d+)\.html$/) {
 	    delete $weirdfiles{$1} if exists $weirdfiles{$1} and -d $1;
+	    delete $weirdfiles{"$1.mbox"}
+	        if exists $weirdfiles{"$1.mbox"} and -f "$1.mbox";
 	}
     }
 
@@ -1452,7 +1679,7 @@ sub prunecache {
     # We now remove the oldfiles if they're automatically downloaded
     tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	 O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
-	or die "bts: couldn't open DB file $timestampdb for writing: $!"
+	or die "bts: couldn't open DB file $timestampdb for writing: $!\n"
 	if ! tied %timestamp;
 
     my @unrecognised;
@@ -1471,7 +1698,7 @@ sub prunecache {
     untie %timestamp;
 
     if (! -e $prunestamp) {
-	open PRUNESTAMP, ">$prunestamp" || die "bts: prune timestamp: $!";
+	open PRUNESTAMP, ">$prunestamp" || die "bts: prune timestamp: $!\n";
 	close PRUNESTAMP;
     }
     utime time, time, $prunestamp;
@@ -1486,6 +1713,21 @@ sub runbrowser {
     }
 }
 
+# Determines which mailreader to use
+sub runmailreader {
+    my $file = shift;
+    die "bts: could not read mbox file!\n" unless -r $file;
+
+    my @reader = split ' ', $mailreader;
+    foreach (@reader) {
+	s/\%([%s])/$1 eq '%' ? '%' : $file/eg;
+    }
+
+    if (system(@reader) >> 8 != 0) {
+	warn "Problem running mail reader: $!\n";
+    }
+}
+
 # Timestamp handling
 # 
 # We store a +ve timestamp to represent an automatic download and
@@ -1501,7 +1743,7 @@ sub get_timestamp {
     } else {
 	tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	     O_RDONLY(), 0600, $DB_HASH, "read")
-	    or die "bts: couldn't open DB file $timestampdb for reading: $!";
+	    or die "bts: couldn't open DB file $timestampdb for reading: $!\n";
 
 	$timestamp = abs($timestamp{$thing})
 	    if exists $timestamp{$thing};
@@ -1521,7 +1763,7 @@ sub set_timestamp {
     } else {
 	tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	     O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
-	    or die "bts: couldn't open DB file $timestampdb for writing: $!";
+	    or die "bts: couldn't open DB file $timestampdb for writing: $!\n";
 
 	$timestamp{$thing} = $timestamp;
 
@@ -1537,7 +1779,7 @@ sub delete_timestamp {
     } else {
 	tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	     O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
-	    or die "bts: couldn't open DB file $timestampdb for writing: $!";
+	    or die "bts: couldn't open DB file $timestampdb for writing: $!\n";
 
 	delete $timestamp{$thing};
 
@@ -1625,13 +1867,13 @@ Converting....
 
 EOW
 
-    chdir $oldcachedir or die "bts: chdir $oldcachedir: $!";
-    opendir DIR, $oldcachedir or die "bts: opendir $oldcachedir: $!";
+    chdir $oldcachedir or die "bts: chdir $oldcachedir: $!\n";
+    opendir DIR, $oldcachedir or die "bts: opendir $oldcachedir: $!\n";
     my @cachefiles = grep { -f $_ } readdir(DIR);
     closedir DIR;
 
-    chdir $oldorigdir or die "bts: chdir $oldorigdir: $!";
-    opendir DIR, $oldorigdir or die "bts: opendir $oldorigdir: $!";
+    chdir $oldorigdir or die "bts: chdir $oldorigdir: $!\n";
+    opendir DIR, $oldorigdir or die "bts: opendir $oldorigdir: $!\n";
     my @origfiles = grep { -f $_ } readdir(DIR);
     closedir DIR;
     my %manual = map { $_ => 1 } grep { s/\.manual$/.html/ } readdir (DIR);
@@ -1669,14 +1911,14 @@ EOW
     # to have a new one ;-)
     if (! -d dirname($cachedir)) {
 	mkdir(dirname($cachedir))
-	    or die "bts: couldn't mkdir ".dirname($cachedir).": $!";
+	    or die "bts: couldn't mkdir ".dirname($cachedir).": $!\n";
     }
     mkdir($cachedir)
-	or die "bts: couldn't mkdir $cachedir: $!";
+	or die "bts: couldn't mkdir $cachedir: $!\n";
 
     tie (%timestamp, "Devscripts::DB_File_Lock", $timestampdb,
 	 O_RDWR()|O_CREAT(), 0600, $DB_HASH, "write")
-	or die "bts: couldn't open DB file $timestampdb for writing: $!";
+	or die "bts: couldn't open DB file $timestampdb for writing: $!\n";
 
     foreach (keys %save_cache_stamps) {
 	my $thing = $_;
@@ -1732,10 +1974,10 @@ EOW
 #                       MIRROR_UP_TO_DATE   up-to-date
 
 sub bts_mirror {
-    my ($url, $timestamp) = @_;
+    my ($url, $timestamp, $force) = @_;
 
     init_agent() unless $ua;
-    if ($url =~ m%/\d+$% and not $refreshmode) {
+    if ($url =~ m%/\d+$% and ! $refreshmode and ! $force) {
 	# Single bug, worth doing timestamp checks
 	my $request = HTTP::Request->new('HEAD', $url);
 	my $response = $ua->request($request);
@@ -1837,16 +2079,17 @@ line parameter being used.  Only has an effect on the show and bug
 commands.  The default is I<yes>.  Again, see the show command above
 for more information.
 
-=item BTS_FULL_MIRROR
+=item BTS_CACHE_MODE={min,mbox,full}
 
-If this is set to I<yes>, then it is the same as the --full-mirror
-command line parameter being used.  Only has an effect on the cache
-command.  The default is I<no>.  See the cache command for more
+How much of the BTS should we mirror when we are asked to cache something?
+Just the minimum, or also the mbox or the whole thing?  The default is
+I<min>, and it has the same meaning as the --cache-mode command line
+parameter.  Only has an effect on the cache.  See the cache command for more
 information.
 
 =item BTS_FORCE_REFRESH
 
-If this is set to I<yes>, then it is the same as the --full-mirror
+If this is set to I<yes>, then it is the same as the --force-refresh
 command line parameter being used.  Only has an effect on the cache
 command.  The default is I<no>.  See the cache command for more
 information.
@@ -1856,7 +2099,7 @@ information.
 =head1 COPYRIGHT
 
 This program is Copyright (C) 2001-2003 by Joey Hess <joeyh@debian.org>.
-Many modifications have been made, Copyright (C) 2002-2004 Julian
+Many modifications have been made, Copyright (C) 2002-2005 Julian
 Gilbey <jdg@debian.org>.
 
 It is licensed under the terms of the GPL, either version 2 of the
