@@ -1,10 +1,12 @@
 #!/bin/bash -e
 
-if test $# -gt 0; then
+PROGNAME=`basename $0`
+
+if [ $# -gt 0 ]; then
     case $1 in
 	-h|--help)
 	    cat <<EOF
-Usage: $0 [options] [<arg> ...]
+Usage: $PROGNAME [options] [<arg> ...]
 Build package and generate build dependencies.
 All args are passed to dpkg-buildpackage.
 Options:
@@ -21,12 +23,20 @@ EOF
     esac
 fi
 
-if test -d debian -a -x debian/rules; then
+if [ -d debian -a -x debian/rules ]; then
     :
 else
     echo "dpkg-genbuilddeps must be run in the source package directory" >&2
     exit 1
 fi
 
+if ! dpkg -L build-essential >/dev/null 2>&1
+then
+    echo "You must have the build-essential package installed to use dpkg-genbuilddeps" >&2
+    echo "You can try running the dpkg-depcheck program directly as:" >&2
+    echo "dpkg-depcheck --all dpkg-buildpackage -us -uc -b -rfakeroot $*" >&2
+    exit 1
+fi
+
 echo "Warning: if this program hangs, kill it and read the manpage!" >&2
 dpkg-depcheck -b dpkg-buildpackage -us -uc -b "$@"
